@@ -16,7 +16,7 @@ from django.utils.unittest.case import skipUnless
 from models import ChangeLog
 import settings as synchro_settings
 from signals import DisableSynchroLog, disable_synchro_log
-from utility import natural_manager, NaturalManager, reset_synchro
+from utility import NaturalManager, reset_synchro, NaturalKeyModel
 
 
 LOCAL = 'default'
@@ -24,7 +24,8 @@ REMOTE = settings.SYNCHRO_REMOTE
 # List of test models
 SETTINGS = {
     'SYNCHRO_MODELS': (
-        ('synchro', 'testmodel', 'PkModelWithSkip', 'ModelWithKey', 'ModelWithFK', 'A', 'X'),
+        ('synchro', 'testmodel', 'PkModelWithSkip', 'ModelWithKey', 'ModelWithFK', 'A', 'X',
+         'M2mModelWithKey', 'M2mAnother', 'M2mModelWithInter'),
     )
 }
 
@@ -89,15 +90,46 @@ class CustomManager(models.Manager):
         return 'Not a single object!'
 
 
-class ModelWithKey(models.Model):
+class ModelWithKey(NaturalKeyModel):
     name = models.CharField(max_length=10)
     cash = models.IntegerField(default=0)
     visits = models.PositiveIntegerField(default=0)
     SYNCHRO_SKIP = ('visits',)
-    objects = natural_manager('name', manager=CustomManager)
+    _natural_key = ('name',)
+
+    objects = CustomManager()
+
+
+class M2mModelWithKey(models.Model):
+    foo = models.IntegerField(default=1)
+    objects = NaturalManager('foo')
 
     def natural_key(self):
-        return self.name,
+        return self.foo,
+
+
+class M2mAnother(models.Model):
+    bar = models.IntegerField(default=1)
+    m2m = models.ManyToManyField('M2mModelWithKey', related_name='r_m2m')
+
+
+class M2mModelWithInter(models.Model):
+    bar = models.IntegerField(default=1)
+    m2m = models.ManyToManyField('M2mModelWithKey', related_name='r_m2m_i',
+                                 through='M2mIntermediate')
+
+
+class M2mNotExplicitlySynced(models.Model):
+    # This model is not listed in SYNCHRO_MODELS
+    foo = models.IntegerField(default=1)
+
+
+class M2mIntermediate(models.Model):
+    with_key = models.ForeignKey(M2mModelWithKey)
+    with_inter = models.ForeignKey(M2mModelWithInter)
+    # To get everything worse, use another FK here, in order to test intermediate sync.
+    extra = models.ForeignKey(M2mNotExplicitlySynced)
+    cash = models.IntegerField()
 
 
 class A(models.Model):
@@ -423,7 +455,7 @@ class AdvancedSynchroTests(SynchroTests):
     """Cover additional features."""
 
     def test_manager_class(self):
-        """Test if natural_manager works."""
+        """Test if NaturalManager works."""
         self.assertIsInstance(ModelWithKey.objects, NaturalManager)
         # Test if it subclasses user manager as well
         self.assertIsInstance(ModelWithKey.objects, CustomManager)
@@ -440,9 +472,16 @@ class AdvancedSynchroTests(SynchroTests):
 
             class X(models.Model):
                 x = models.IntegerField()
-                objects = natural_manager('x', manager=BadManager)
+                objects = NaturalManager('x', manager=BadManager)
         self.assertRaises(ValidationError, wrong)  # User manager must subclass Manager
 
+        # Test if manager without fields raises exception
+        def wrong2():
+            class X(models.Model):
+                x = models.IntegerField()
+                objects = NaturalManager()
+        self.assertRaises(AssertionError, wrong2)
+
     def test_natural_key(self):
         """
         Test if natural key works.
@@ -572,3 +611,141 @@ class SignalSynchroTests(SynchroTests):
     def test_signals_alternative_good_scenario(self):
         """Demonstrate solution for scenario from README (use update)."""
         self._test_signals_scenario(update_bar_good_upd, 42)  # GOOD RESULT
+
+
+class M2MSynchroTests(SynchroTests):
+    """Cover many2many relations tests."""
+
+    def test_natural_manager(self):
+        """Test if natural manager can be instantiated when using M2M."""
+        test = M2mModelWithKey.objects.create()
+        obj = M2mAnother.objects.create()
+        obj.m2m.add(test)  # this would fail if NaturalManager could not be instantiated
+        self.assertEqual(test.pk, obj.m2m.get_by_natural_key(1).pk)
+
+    def test_simple_m2m(self):
+        """Test if m2m field is synced properly."""
+        test = M2mModelWithKey.objects.create()
+        a = M2mAnother.objects.create()
+
+        # add
+        a.m2m.add(test)
+        self.synchronize()
+        self.assertRemoteCount(1, M2mAnother)
+        self.assertRemoteCount(1, M2mModelWithKey)
+        b = M2mAnother.objects.db_manager(REMOTE).all()[0]
+        k = M2mModelWithKey.objects.db_manager(REMOTE).all()[0]
+        self.assertEqual(1, b.m2m.count())
+        self.assertEqual(1, k.r_m2m.count())
+        b_k = b.m2m.all()[0]
+        self.assertEqual(b_k.pk, k.pk)
+        self.assertEqual(b_k.foo, k.foo)
+
+        # clear
+        self.wait()
+        a.m2m.clear()
+        self.synchronize()
+        self.assertEqual(0, b.m2m.count())
+        self.assertEqual(0, k.r_m2m.count())
+
+        # reverse add
+        self.wait()
+        a2 = M2mAnother.objects.create(bar=2)
+        test.r_m2m.add(a, a2)
+        self.synchronize()
+        self.assertRemoteCount(2, M2mAnother)
+        self.assertRemoteCount(1, M2mModelWithKey)
+        b2 = M2mAnother.objects.db_manager(REMOTE).filter(bar=2)[0]
+        self.assertEqual(1, b.m2m.count())
+        self.assertEqual(1, b2.m2m.count())
+        self.assertEqual(2, k.r_m2m.count())
+
+        # reverse remove
+        self.wait()
+        test.r_m2m.remove(a)
+        self.synchronize()
+        self.assertRemoteCount(2, M2mAnother)
+        self.assertRemoteCount(1, M2mModelWithKey)
+        self.assertEqual(0, b.m2m.count())
+        self.assertEqual(1, b2.m2m.count())
+        self.assertEqual(1, k.r_m2m.count())
+
+        # reverse clear
+        self.wait()
+        test.r_m2m.clear()
+        self.synchronize()
+        self.assertRemoteCount(2, M2mAnother)
+        self.assertRemoteCount(1, M2mModelWithKey)
+        self.assertEqual(0, b.m2m.count())
+        self.assertEqual(0, b2.m2m.count())
+        self.assertEqual(0, k.r_m2m.count())
+
+    def test_intermediary_m2m(self):
+        """Test if m2m field with explicit intermediary is synced properly."""
+        test = M2mNotExplicitlySynced.objects.create(foo=77)
+        key = M2mModelWithKey.objects.create()
+        a = M2mModelWithInter.objects.create()
+        M2mIntermediate.objects.create(with_key=key, with_inter=a, cash=42, extra=test)
+        self.assertEqual(1, a.m2m.count())
+        self.assertEqual(1, key.r_m2m_i.count())
+        self.synchronize()
+        self.assertRemoteCount(1, M2mNotExplicitlySynced)
+        self.assertRemoteCount(1, M2mModelWithKey)
+        self.assertRemoteCount(1, M2mModelWithInter)
+        b = M2mModelWithInter.objects.db_manager(REMOTE).all()[0]
+        k = M2mModelWithKey.objects.db_manager(REMOTE).all()[0]
+        self.assertEqual(1, b.m2m.count())
+        self.assertEqual(1, k.r_m2m_i.count())
+        b_k = b.m2m.all()[0]
+        self.assertEqual(b_k.pk, k.pk)
+        self.assertEqual(b_k.foo, k.foo)
+        self.assertEqual(1, b.m2m.all()[0].foo)
+        # intermediary
+        self.assertRemoteCount(1, M2mIntermediate)
+        inter = M2mIntermediate.objects.db_manager(REMOTE).all()[0]
+        self.assertEqual(42, inter.cash)
+        self.assertEqual(77, inter.extra.foo)  # check if extra FK model get synced
+
+        # changing
+        self.wait()
+        key2 = M2mModelWithKey.objects.create(foo=42)
+        a.m2m.clear()
+        M2mIntermediate.objects.create(with_key=key2, with_inter=a, cash=77, extra=test)
+        self.synchronize()
+        self.assertRemoteCount(1, M2mNotExplicitlySynced)
+        self.assertRemoteCount(2, M2mModelWithKey)
+        self.assertRemoteCount(1, M2mModelWithInter)
+        b = M2mModelWithInter.objects.db_manager(REMOTE).all()[0]
+        self.assertEqual(1, b.m2m.count())
+        self.assertEqual(42, b.m2m.all()[0].foo)
+        # intermediary
+        self.assertRemoteCount(1, M2mIntermediate)
+        inter = M2mIntermediate.objects.db_manager(REMOTE).all()[0]
+        self.assertEqual(77, inter.cash)
+
+        # intermadiate change
+        self.wait()
+        inter = M2mIntermediate.objects.all()[0]
+        inter.cash = 1
+        inter.save()
+        self.synchronize()
+        # No changes here
+        self.assertRemoteCount(1, M2mNotExplicitlySynced)
+        self.assertRemoteCount(2, M2mModelWithKey)
+        self.assertRemoteCount(1, M2mModelWithInter)
+        b = M2mModelWithInter.objects.db_manager(REMOTE).all()[0]
+        self.assertEqual(1, b.m2m.count())
+        self.assertEqual(42, b.m2m.all()[0].foo)
+        # Still one intermediary
+        self.assertRemoteCount(1, M2mIntermediate)
+        inter = M2mIntermediate.objects.db_manager(REMOTE).all()[0]
+        self.assertEqual(1, inter.cash)
+
+
+        # Tricky: clear from other side of relation.
+        self.wait()
+        key2.r_m2m_i.clear()
+        self.synchronize()
+        b = M2mModelWithInter.objects.db_manager(REMOTE).all()[0]
+        self.assertEqual(0, b.m2m.count())
+        self.assertRemoteCount(0, M2mIntermediate)
